@@ -14,8 +14,10 @@ extern int errno;
 
 #if	MSDOS
 #include "unixemu.h"
-#else
+#else	/* !MSDOS */
 #include <sys/time.h>
+#include <sys/param.h>
+#include <sys/file.h>
 # ifdef	USEDIRECT
 # include <sys/dir.h>
 #  ifdef	DIRSIZ
@@ -27,12 +29,16 @@ extern int errno;
 # ifdef	USEUTIME
 # include <utime.h>
 # endif
-#endif
+#endif	/* !MSDOS */
 
 #ifdef	USETIMEH
 #include <time.h>
 #endif
 
+#ifndef	_NODOSDRIVE
+#include "dosdisk.h"
+#endif
+
 #if	(GETTODARGS == 1)
 #define	gettimeofday2(tv, tz)	gettimeofday(tv)
 #else
@@ -44,21 +50,45 @@ extern int errno;
 #define	b_realloc(ptr, n, type) \
 			(((n) % BUFUNIT) ? ((type *)(ptr)) \
 			: (type *)realloc2(ptr, b_size(n, type)))
-#define	c_malloc(size)	(malloc2((size) = BUFUNIT))
-#define	c_realloc(ptr, n, size) \
-			(((n) + 1 < (size)) \
-			? (ptr) : realloc2(ptr, (size) *= 2))
 
 #define	getblock(c)	((((c) + blocksize - 1) / blocksize) * blocksize)
 
+#define	isleftshift()	(n_column < WCOLUMNSTD)
+#define	isrightomit()	(n_column < WCOLUMNSTD - 1)
+#define	ispureshift()	(n_column == WCOLUMNSTD - 1)
+#define	isshortwid()	(n_column < WCOLUMNSTD - 1)
+#define	iswellomit()	(n_column < WCOLUMNSTD - 2)
+#define	ishardomit()	(n_column < WCOLUMNOMIT)
+#define	isbestomit()	(n_column < WCOLUMNHARD)
+
+#ifdef	_NOORIGSHELL
+#define	getconstvar(s)		(char *)getenv(s)
+#define	getshellvar(s, l)	(char *)getenv(s)
+#else
+#define	getconstvar(s)		(getshellvar(s, sizeof(s) - 1))
+#endif
+
+#ifdef	USESTRERROR
+#define	strerror2		strerror
+#else
+# ifndef	DECLERRLIST
+extern char *sys_errlist[];
+# endif
+#define	strerror2(n)		(char *)sys_errlist[n]
+#endif
+
 /* main.c */
 extern VOID error __P_((char *));
+extern VOID checkscreen __P_((int, int));
 extern int sigvecset __P_((int));
-extern VOID title __P_((void));
+extern VOID title __P_((VOID_A));
+#ifndef	_NOCUSTOMIZE
+extern VOID saveorigenviron __P_((VOID_A));
+#endif
 extern int loadruncom __P_((char *, int));
 
 /* dosemu.c or unixemu.c */
-#if	MSDOS || !defined (_NODOSDRIVE)
+#ifdef	_USEDOSPATH
 extern int _dospath __P_((char *));
 extern int dospath __P_((char *, char *));
 #endif
@@ -71,15 +101,7 @@ extern int dospath3 __P_((char *));
 #define	dospath3(path)	dospath(path, NULL)
 # endif
 #endif
-#if	!MSDOS && !defined (_NOKANJIFCONV)
-extern int getkcode __P_((char *));
-#endif
-extern DIR *_Xopendir __P_((char *));
-#ifdef	_NOROCKRIDGE
-#define	Xopendir	_Xopendir
-#else
 extern DIR *Xopendir __P_((char *));
-#endif
 extern int Xclosedir __P_((DIR *));
 extern struct dirent *Xreaddir __P_((DIR *));
 extern VOID Xrewinddir __P_((DIR *));
@@ -88,24 +110,9 @@ extern int rawchdir __P_((char *));
 #else
 #define	rawchdir	chdir
 #endif
-#if	MSDOS && defined (_NODOSDRIVE)
-#define	_Xchdir		rawchdir
-#else
-extern int _Xchdir __P_((char *));
-#endif
-#ifdef	_NOROCKRIDGE
-#define	Xchdir		_Xchdir
-#else
 extern int Xchdir __P_((char *));
-#endif
-extern char *_Xgetwd __P_((char *));
 extern char *Xgetwd __P_((char *));
 extern int Xstat __P_((char *, struct stat *));
-#if	MSDOS || defined (_NOKANJIFCONV)
-#define	_Xlstat(p, s)	(lstat(p, s) ? -1 : 0)
-#else
-extern int _Xlstat __P_((char *, struct stat *));
-#endif
 extern int Xlstat __P_((char *, struct stat *));
 extern int Xaccess __P_((char *, int));
 extern int Xsymlink __P_((char *, char *));
@@ -118,11 +125,7 @@ extern int Xutimes __P_((char *, struct timeval []));
 #endif
 extern int Xunlink __P_((char *));
 extern int Xrename __P_((char *, char *));
-#if	MSDOS && defined (_NOROCKRIDGE) && defined (_NOUSELFN)
-#define	Xopen		open
-#else
 extern int Xopen __P_((char *, int, int));
-#endif
 #ifdef	_NODOSDRIVE
 #define	Xclose		close
 #define	Xread		read
@@ -141,54 +144,9 @@ extern off_t Xlseek __P_((int, off_t, int));
 extern int Xdup __P_((int));
 extern int Xdup2 __P_((int, int));
 #endif
-#if	MSDOS && defined (_NOUSELFN)
-# ifdef	DJGPP
-#define	_Xmkdir(p, m)	(mkdir(p, m) ? -1 : 0)
-# else
-extern int _Xmkdir __P_((char *, int));
-# endif
-#define	_Xrmdir(p)	(rmdir(p) ? -1 : 0)
-#else
-# if	!MSDOS && defined (_NODOSDRIVE) && defined (_NOKANJIFCONV)
-#define	_Xmkdir		mkdir
-#define	_Xrmdir		rmdir
-# else
-extern int _Xmkdir __P_((char *, int));
-extern int _Xrmdir __P_((char *));
-# endif
-#endif
-#ifdef	_NOROCKRIDGE
-# if	MSDOS && defined (_NOUSELFN)
-#  ifdef	DJGPP
-#define	Xmkdir(p, m)	(mkdir(p, m) ? -1 : 0)
-#  else
-#define	Xmkdir(p, m)	(mkdir(p) ? -1 : 0)
-#  endif
-#define	Xrmdir(p)	(rmdir(p) ? -1 : 0)
-# else
-#  if	!MSDOS && defined (_NODOSDRIVE) && defined (_NOKANJIFCONV)
-#define	Xmkdir		mkdir
-#define	Xrmdir		rmdir
-#  else
-#define	Xmkdir		_Xmkdir
-#define	Xrmdir		_Xrmdir
-#  endif
-# endif
-#else
 extern int Xmkdir __P_((char *, int));
 extern int Xrmdir __P_((char *));
-#endif
-#if	(MSDOS && defined (_NOUSELFN)) \
-|| (!MSDOS && defined (_NODOSDRIVE) && defined (_NOKANJIFCONV))
-#define	_Xfopen		fopen
-#else
-extern FILE *_Xfopen __P_((char *, char *));
-#endif
-#ifdef	_NOROCKRIDGE
-#define	Xfopen		_Xfopen
-#else
 extern FILE *Xfopen __P_((char *, char *));
-#endif
 #ifdef	_NODOSDRIVE
 #define	Xfdopen		fdopen
 #define	Xfclose		fclose
@@ -232,40 +190,47 @@ extern int mkdir2 __P_((char *, int));
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
 extern char *strdup2 __P_((char *));
+extern char *strndup2 __P_((char *, int));
+extern char *c_realloc __P_((char *, ALLOC_T, ALLOC_T *));
 extern char *strchr2 __P_((char *, int));
 extern char *strrchr2 __P_((char *, int));
 extern char *strcpy2 __P_((char *, char *));
 extern char *strncpy2 __P_((char *, char *, int));
 extern int strncpy3 __P_((char *, char *, int *, int));
-extern char *strstr2 __P_((char *, char *));
 extern int strlen2 __P_((char *));
 extern int strlen3 __P_((char *));
 extern int atoi2 __P_((char *));
-extern int putenv2 __P_((char *));
-extern int setenv2 __P_((char *, char *));
+extern int fprintf2 __P_((FILE *, CONST char *, ...));
+extern int snprintf2 __P_((char *, int, CONST char *, ...));
+extern VOID perror2 __P_((char *));
+extern int setenv2 __P_((char *, char *, int));
 extern char *getenv2 __P_((char *));
-#ifdef	DEBUG
-extern VOID freeenv __P_((VOID_A));
+#ifdef	USESIGACTION
+extern sigcst_t signal2 __P_((int, sigcst_t));
+#else
+#define	signal2	signal
 #endif
 extern int system2 __P_((char *, int));
+extern FILE *popen2 __P_((char *, char *));
+extern int pclose2 __P_((FILE *));
 extern char *getwd2 __P_((VOID_A));
+extern time_t time2 __P_((VOID_A));
 extern time_t timelocal2 __P_((struct tm *));
 extern char *fgets2 __P_((FILE *, int));
 
 /* file.c */
-#if	MSDOS || defined (_NODOSDRIVE)
-#define	nodospath(p, f)	(f)
+#ifdef	_USEDOSEMU
+extern char *nodospath __P_((char *, char *));
 #else
-char *nodospath __P_((char *, char *));
+#define	nodospath(p, f)	(f)
 #endif
 #define	fnodospath(p, i)	nodospath(p, filelist[i].name)
-#if	MSDOS
-extern int logical_access __P_((u_short));
+#ifdef	NOUID
+extern int logical_access __P_((u_int));
 #else
-extern int logical_access __P_((u_short, uid_t, gid_t));
+extern int logical_access __P_((u_int, uid_t, gid_t));
 #endif
 extern int getstatus __P_((namelist *));
-extern int isdotdir __P_((char *));
 extern int cmplist __P_((CONST VOID_P, CONST VOID_P));
 #ifndef	_NOTREE
 extern int cmptree __P_((CONST VOID_P, CONST VOID_P));
@@ -274,21 +239,27 @@ extern struct dirent *searchdir __P_((DIR *, reg_t *, char *));
 extern int underhome __P_((char *));
 extern int preparedir __P_((char *));
 extern int touchfile __P_((char *, struct stat *));
+#ifdef	_NODOSDRIVE
+#define	nodoslstat	Xlstat
+#else
+extern int nodoslstat __P_((char *, struct stat *));
+#endif
 extern VOID lostcwd __P_((char *));
 extern int safewrite __P_((int, char *, int));
 extern int safecpfile __P_((char *, char *, struct stat *, struct stat *));
 extern int safemvfile __P_((char *, char *, struct stat *, struct stat *));
+extern char *genrandname __P_((char *, int));
 extern int mktmpdir __P_((char *));
 extern int rmtmpdir __P_((char *));
-extern int mktmpfile __P_((char *, char *));
+extern int mktmpfile __P_((char *));
 extern int rmtmpfile __P_((char *));
-extern VOID removetmp __P_((char *, char *, char *));
+extern VOID removetmp __P_((char *, char *));
 extern int forcecleandir __P_((char *, char *));
 #ifndef	_NODOSDRIVE
+extern char *dostmpdir __P_((int));
 extern int tmpdosdupl __P_((char *, char **, int));
 extern int tmpdosrestore __P_((int, char *));
 #endif
-int isexist __P_((char *));
 #ifndef	_NOWRITEFS
 extern VOID arrangedir __P_((int));
 #endif
@@ -298,18 +269,19 @@ extern int rmvfile __P_((char *));
 extern int rmvdir __P_((char *));
 extern int findfile __P_((char *));
 extern int finddir __P_((char *));
-extern int inputattr __P_((namelist *, u_short));
+extern int inputattr __P_((namelist *, int));
 extern int setattr __P_((char *));
 extern int applyfile __P_((int (*)__P_((char *)), char *));
 extern int applydir __P_((char *, int (*)__P_((char *)),
 		int (*)__P_((char *)), int (*)__P_((char *)), int, char *));
 extern int copyfile __P_((char *, int));
 extern int movefile __P_((char *, int));
+extern int forcemovefile __P_((char *));
 
 /* parse.c */
 extern char *skipspace __P_((char *));
 extern char *evalnumeric __P_((char *, long *, int));
-extern char *ascnumeric __P_((char *, long, int, int));
+extern char *ascnumeric __P_((char *, off_t, int, int));
 #ifdef	_NOORIGSHELL
 extern char *strtkchr __P_((char *, int, int));
 extern int getargs __P_((char *, char ***));
@@ -323,35 +295,45 @@ extern char *evalpaths __P_((char *, int));
 #else
 extern char *killmeta __P_((char *));
 #endif
-#if	!MSDOS && defined (_NOORIGSHELL)
+#ifdef	_NOORIGSHELL
 extern VOID adjustpath __P_((VOID_A));
 #endif
-extern char *includepath __P_((char *, char *, char **));
-extern VOID freevar __P_((char **));
+extern char *includepath __P_((char *, char *));
 #if	(FD < 2) && !defined (_NOARCHIVE)
 extern char *getrange __P_((char *, u_char *, u_char *, u_char *));
 #endif
-extern int evalprompt __P_((char *, char *, int));
+extern int evalprompt __P_((char **, char *));
+#if	FD >= 2
+extern char **duplvar __P_((char **, int));
+#endif
 #ifndef	_NOARCHIVE
-extern char *getext __P_((char *, int *));
+extern char *getext __P_((char *, u_char *));
 extern int extcmp __P_((char *, int, char *, int, int));
 #endif
 extern int getkeycode __P_((char *, int));
 extern char *getkeysym __P_((int, int));
 extern char *decodestr __P_((char *, int *, int));
-#if	!MSDOS && !defined (_NOKEYMAP)
+#ifndef	_NOKEYMAP
 extern char *encodestr __P_((char *, int));
 #endif
 
 /* builtin.c */
+#ifndef	_NOARCHIVE
 extern VOID printlaunchcomm __P_((int, int, FILE *));
 extern VOID printarchcomm __P_((int, FILE *));
+# ifndef	_NOBROWSE
+extern VOID freebrowse __P_((launchtable *));
+# endif
+#endif
 extern int freemacro __P_((int));
 extern VOID printmacro __P_((int, FILE *));
+#ifdef	_USEDOSEMU
+extern int searchdrv __P_((devinfo *, int, char *, int, int, int, int));
 extern int deletedrv __P_((int));
 extern int insertdrv __P_((int, int, char *, int, int, int));
 extern VOID printsetdrv __P_((int, int, FILE *));
-#if	!MSDOS && !defined (_NOKEYMAP)
+#endif
+#ifndef	_NOKEYMAP
 extern VOID printkeymap __P_((int, char *, int, FILE *));
 #endif
 #ifdef	_NOORIGSHELL
@@ -374,9 +356,11 @@ extern VOID freedefine __P_((VOID_A));
 #endif
 
 #define	BL_SET		"set"
+#define	BL_UNSET	"unset"
 #define	BL_PENV		"printenv"
 #define	BL_LAUNCH	"launch"
 #define	BL_ARCH		"arch"
+#define	BL_BROWSE	"browse"
 #define	BL_PLAUNCH	"printlaunch"
 #define	BL_PARCH	"printarch"
 #define	BL_BIND		"bind"
@@ -389,8 +373,10 @@ extern VOID freedefine __P_((VOID_A));
 #define	BL_HISTORY	"history"
 #define	BL_FC		"fc"
 #define	BL_CHECKID	"checkid"
-#define	BL_KCONV	"kconv"
 #define	BL_EVALMACRO	"evalmacro"
+#define	BL_KCONV	"kconv"
+#define	BL_READLINE	"readline"
+#define	BL_YESNO	"yesno"
 #define	BL_ALIAS	"alias"
 #define	BL_UALIAS	"unalias"
 #define	BL_FUNCTION	"function"
@@ -402,9 +388,15 @@ extern VOID freedefine __P_((VOID_A));
 
 /* shell.c */
 extern char *evalcommand __P_((char *, char *, macrostat *, int));
-extern char *inputshellstr __P_((char *, int, int, char *));
+extern int replaceargs __P_((int *, char ***, char **, int));
+extern int replacearg __P_((char **));
+extern int argfputs __P_((char *, FILE *));
+extern VOID demacroarg __P_((char **));
+extern char *inputshellstr __P_((char *, int, char *));
 extern char *inputshellloop __P_((int, char *));
+extern int isinternalcomm __P_((char *));
 extern int execmacro __P_((char *, char *, int, int, int));
+extern FILE *popenmacro __P_((char *, char *, int));
 #ifdef	_NOORIGSHELL
 extern int execusercomm __P_((char *, char *, int, int, int));
 #else
@@ -426,37 +418,40 @@ extern int completeuserfunc __P_((char *, int, int, char ***));
 
 /* kanji.c */
 extern int onkanji1 __P_((char *, int));
-#if	(!MSDOS && !defined (_NOKANJICONV)) \
+#if	!defined (_NOKANJICONV) \
 || (!defined (_NOENGMES) && !defined (_NOJPNMES))
 extern int getlang __P_((char *, int));
 #endif
-#if	!MSDOS && (!defined (_NOKANJICONV) \
-|| (!defined (_NODOSDRIVE) && defined (CODEEUC)))
+#if	!defined (_NOKANJICONV) \
+|| (defined (_USEDOSEMU) && defined (CODEEUC))
 extern int sjis2ujis __P_((char *, u_char *, int));
 extern int ujis2sjis __P_((char *, u_char *, int));
 #endif
-#if	(!MSDOS && defined (FD) && (FD >= 2) && !defined (_NOKANJICONV)) \
-|| !defined (_NODOSDRIVE)
+#if	!defined (_NOKANJICONV) || !defined (_NODOSDRIVE)
 extern VOID readunitable __P_((VOID_A));
 extern VOID discardunitable __P_((VOID_A));
-extern u_short unifysjis __P_((u_short, int));
-extern u_short cnvunicode __P_((u_short, int));
+extern u_int unifysjis __P_((u_int, int));
+extern u_int cnvunicode __P_((u_int, int));
 #endif
-#if	!MSDOS && !defined (_NOKANJICONV)
+#ifndef	_NOKANJICONV
 extern int kanjiconv __P_((char *, char *, int, int, int, int));
 # ifndef	_NOKANJIFCONV
 extern char *kanjiconv2 __P_((char *, char *, int, int, int));
 # endif
 #endif
-extern int kanjiputs __P_((char *));
 extern int kanjifputs __P_((char *, FILE *));
+extern int kanjiputs __P_((char *));
 extern int kanjiprintf __P_((CONST char *, ...));
 extern int kanjiputs2 __P_((char *, int, int));
+#ifndef	_NOKANJIFCONV
+extern int getkcode __P_((char *));
+#endif
+extern char *convget __P_((char *, char *, int));
+extern char *convput __P_((char *, char *, int, int, char *, int *));
 
 /* input.c */
 extern int intrkey __P_((VOID_A));
 extern int Xgetkey __P_((int, int));
-extern int cmdlinelen __P_((int));
 extern char *inputstr __P_((char *, int, int, char *, int));
 extern int yesno __P_((CONST char *, ...));
 extern VOID warning __P_((int, char *));
@@ -465,20 +460,22 @@ extern int selectstr __P_((int *, int, int, char *[], int []));
 /* info.c */
 extern VOID help __P_((int));
 extern int writablefs __P_((char *));
-extern long getblocksize __P_((char *));
-extern VOID getinfofs __P_((char *, long *, long *));
+extern off_t getblocksize __P_((char *));
+extern off_t calcKB __P_((off_t, off_t));
+extern int getinfofs __P_((char *, off_t *, off_t *, off_t *));
 extern int infofs __P_((char *));
 
 /* rockridge.c */
 #ifndef	_NOROCKRIDGE
-extern int transfilelist __P_((VOID_A));
-extern char *transfile __P_((char *, char *));
-extern char *detransfile __P_((char *, char *, int));
+extern char *transpath __P_((char *, char *));
+extern char *detranspath __P_((char *, char *));
+extern int rrlstat __P_((char *, struct stat *));
+extern int rrreadlink __P_((char *, char *, int));
 #endif
 
 /* archive.c */
 #ifndef	_NOARCHIVE
-extern VOID poparchdupl __P_((VOID_A));
+extern VOID escapearch __P_((VOID_A));
 extern VOID archbar __P_((char *, char *));
 extern VOID copyarcf __P_((reg_t *, char *));
 extern char *archchdir __P_((char *));
@@ -486,6 +483,7 @@ extern char *archchdir __P_((char *));
 extern int completearch __P_((char *, int, int, char ***));
 # endif
 extern int launcher __P_((VOID_A));
+extern int dolaunch __P_((launchtable *, int));
 extern int pack __P_((char *));
 extern int unpack __P_((char *, char *, char *, int, int));
 extern char *tmpunpack __P_((int));
@@ -505,11 +503,20 @@ extern VOID evalenv __P_((VOID_A));
 extern VOID freeenvpath __P_((VOID_A));
 #endif
 #ifndef	_NOCUSTOMIZE
-char **copystrarray __P_((char **, char **, int *, int));
+extern VOID freestrarray __P_((char **, int));
+extern char **copystrarray __P_((char **, char **, int *, int));
 extern bindtable *copybind __P_((bindtable *, bindtable *));
-# if	!MSDOS && !defined (_NOKEYMAP)
-extern keymaptable *copykeymap __P_((keymaptable *));
-extern VOID freekeymap __P_((keymaptable *));
+# ifndef	_NOARCHIVE
+extern VOID freelaunch __P_((launchtable *, int));
+extern launchtable *copylaunch __P_((launchtable *, launchtable *,
+	int *, int));
+extern VOID freearch __P_((archivetable *, int));
+extern archivetable *copyarch __P_((archivetable *, archivetable *,
+	int *, int));
+# endif
+# ifdef	_USEDOSEMU
+extern VOID freedosdrive __P_((devinfo *));
+extern devinfo *copydosdrive __P_((devinfo *, devinfo *));
 # endif
 extern VOID rewritecust __P_((int));
 extern int customize __P_((VOID_A));
@@ -519,9 +526,9 @@ extern int customize __P_((VOID_A));
 
 /* browse.c */
 extern VOID helpbar __P_((VOID_A));
-extern char *putmode __P_((char *, u_short));
+extern int putmode __P_((char *, u_int));
 #ifdef	HAVEFLAGS
-extern char *putflags __P_((char *, u_long));
+extern int putflags __P_((char *, u_long));
 #endif
 extern VOID waitmes __P_((VOID_A));
 extern int calcwidth __P_((VOID_A));
@@ -530,7 +537,7 @@ extern int listupfile __P_((namelist *, int, char *));
 #ifndef	_NOSPLITWIN
 extern int shutwin __P_((int));
 #endif
-extern VOID movepos __P_((int, u_char));
+extern VOID movepos __P_((int, int));
 extern VOID rewritefile __P_((int));
 extern VOID addlist __P_((VOID_A));
 extern VOID main_fd __P_((char *));
