@@ -11,12 +11,6 @@
 #include "kctype.h"
 #include "kanji.h"
 
-#ifdef	_NOORIGSHELL
-#define	dosystem(s)		system(s)
-#else
-#include "system.h"
-#endif
-
 #ifndef	NOTZFILEH
 #include <tzfile.h>
 #endif
@@ -25,12 +19,22 @@
 #include <sys/timeb.h>
 extern int setcurdrv __P_((int, int));
 extern char *unixrealpath __P_((char *, char *));
-#else
-#include <sys/file.h>
-#include <sys/param.h>
 #endif
+
 #ifndef	_NODOSDRIVE
 extern int flushdrv __P_((int, VOID_T (*)__P_((VOID_A))));
+extern char *dosgetcwd __P_((char *, int));
+#endif
+
+#ifdef	_NOORIGSHELL
+#include <signal.h>
+#define	dosystem(s)		system(s)
+#else
+#include "system.h"
+#endif
+
+#ifdef	CYGWIN
+#include <sys/cygwin.h>
 #endif
 
 #ifdef	PWNEEDERROR
@@ -43,7 +47,13 @@ extern char **environ;
 extern char fullpath[];
 extern char *origpath;
 extern int hideclock;
+#ifndef	_NODOSDRIVE
+extern int lastdrv;
+#endif
 
+#ifndef	SIG_ERR
+#define	SIG_ERR		((sigcst_t)-1)
+#endif
 #ifndef	CHAR_BIT
 # ifdef	NBBY
 # define	CHAR_BIT	NBBY
@@ -56,6 +66,9 @@ extern int hideclock;
 			| (((u_char *)(cp))[1] << (CHAR_BIT * 2)) \
 			| (((u_char *)(cp))[0] << (CHAR_BIT * 3)) )
 
+#ifndef	NOSYMLINK
+static int NEAR evallink __P_((char *, char *));
+#endif
 static char *NEAR _realpath2 __P_((char *, char *, int));
 #ifdef	_NOORIGSHELL
 static int NEAR _getenv2 __P_((char *, int, char **));
@@ -72,44 +85,85 @@ static long NEAR gettimezone __P_((struct tm *, time_t));
 #ifdef	_NOORIGSHELL
 char **environ2 = NULL;
 #endif
+int physical_path = 0;
+int norealpath = 0;
 
 static char *lastpath = NULL;
 #ifndef	_NODOSDRIVE
 static char *unixpath = NULL;
 #endif
+static int wasttyflags = 0;
+#define	TF_TTYIOMODE	001
+#define	TF_TTYNL	002
 
 
 int stat2(path, stp)
 char *path;
 struct stat *stp;
 {
-#if	MSDOS
+#ifdef	NOSYMLINK
 	return(Xstat(path, stp));
-#else	/* !MSDOS */
-	int tmperr;
+#else	/* !NOSYMLINK */
+	int duperrno;
 
 	if (Xstat(path, stp) < 0) {
-#ifndef	_NODOSDRIVE
-		if (dospath2(path)) return(-1);
-#endif
-		tmperr = errno;
-		if (_Xlstat(path, stp) < 0
+		duperrno = errno;
+# ifndef	_NODOSDRIVE
+		if (dospath2(path)) {
+			errno = duperrno;
+			return(-1);
+		}
+# endif
+		if (nodoslstat(path, stp) < 0
 		|| (stp -> st_mode & S_IFMT) != S_IFLNK) {
-			errno = tmperr;
+			errno = duperrno;
 			return(-1);
 		}
 		stp -> st_mode &= ~S_IFMT;
 	}
 	return(0);
-#endif	/* !MSDOS */
+#endif	/* !NOSYMLINK */
 }
 
+#ifndef	NOSYMLINK
+static int NEAR evallink(path, delim)
+char *path, *delim;
+{
+	struct stat st;
+	char buf[MAXPATHLEN];
+	int i, duperrno;
+
+	duperrno = errno;
+	if (nodoslstat(path, &st) < 0) {
+		errno = duperrno;
+		return(-1);
+	}
+	if ((st.st_mode & S_IFMT) != S_IFLNK) {
+		errno = duperrno;
+		return(0);
+	}
+	if ((i = Xreadlink(path, buf, MAXPATHLEN - 1)) < 0) {
+		errno = duperrno;
+		return(-1);
+	}
+
+	if (*buf == _SC_ || !delim || delim == path) {
+		*path = _SC_;
+		delim = path + 1;
+	}
+	buf[i] = *delim = '\0';
+	_realpath2(buf, path, 1);
+	errno = duperrno;
+	return(1);
+}
+#endif	/* !NOSYMLINK */
+
 /*ARGSUSED*/
 static char *NEAR _realpath2(path, resolved, rdlink)
 char *path, *resolved;
 int rdlink;
 {
-	char *cp;
+	char *cp, *top;
 
 	if (!*path || !strcmp(path, ".")) return(resolved);
 	else if ((cp = strdelim(path, 0))) {
@@ -122,50 +176,33 @@ int rdlink;
 
 	if (!strcmp(path, "..")) {
 		cp = strrdelim(resolved, 0);
+		top = resolved;
 #if	MSDOS
-		if (cp && cp != &(resolved[2])) *cp = '\0';
-		else resolved[3] = '\0';
+		top += 2;
 #else	/* !MSDOS */
-#ifndef	_NODOSDRIVE
-		if (_dospath(resolved)) {
-			if (cp && cp != &(resolved[2])) *cp = '\0';
-			else resolved[3] = '\0';
-		}
+# ifndef	_NODOSDRIVE
+		if (_dospath(resolved)) top += 2;
 		else
-#endif
-		{
-			if (cp && cp != resolved) *cp = '\0';
-			else resolved[1] = '\0';
-		}
+# endif
+		if (rdlink && evallink(resolved, cp) > 0)
+			return(_realpath2(path, resolved, rdlink));
 #endif	/* !MSDOS */
+		if (!cp || cp == top) {
+			*top = _SC_;
+			cp = top + 1;
+		}
+		*cp = '\0';
 	}
 	else {
 		cp = strcatdelim(resolved);
 		strncpy2(cp, path, MAXPATHLEN - 1 - (cp - resolved));
-#if	!MSDOS
-		if (!rdlink);
+#ifndef	NOSYMLINK
+		if (!rdlink) /*EMPTY*/;
 # ifndef	_NODOSDRIVE
-		else if (_dospath(resolved));
+		else if (_dospath(resolved)) /*EMPTY*/;
 # endif
-		else {
-			struct stat st;
-			char buf[MAXPATHLEN];
-			int i;
-
-			if (_Xlstat(resolved, &st) >= 0
-			&& (st.st_mode & S_IFMT) == S_IFLNK
-			&& (i = readlink(resolved, buf, MAXPATHLEN - 1)) >= 0)
-			{
-				buf[i] = '\0';
-				if (*buf == _SC_) strcpy(resolved, buf);
-				else {
-					if (cp - 1 > resolved) cp--;
-					*cp = '\0';
-					_realpath2(buf, resolved, rdlink);
-				}
-			}
-		}
-#endif
+		else evallink(resolved, cp - 1);
+#endif	/* !NOSYMLINK */
 	}
 	return(resolved);
 }
@@ -174,8 +211,14 @@ char *realpath2(path, resolved, rdlink)
 char *path, *resolved;
 int rdlink;
 {
-	char tmp[MAXPATHLEN];
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	char *cp;
+	int duplastdrive;
+#endif
+#if	MSDOS || !defined (_NODOSDRIVE)
 	int drv;
+#endif
+	char tmp[MAXPATHLEN];
 
 	strcpy(tmp, path);
 	path = tmp;
@@ -195,33 +238,39 @@ int rdlink;
 			resolved[2] = _SC_;
 			resolved[3] = '\0';
 		}
-		else if (!Xgetwd(resolved) || setcurdrv(drv, 0) < 0)
-			error(NULL);
+		else {
+			if (!Xgetwd(resolved)) lostcwd(resolved);
+			if (setcurdrv(drv, 0) < 0) error("setcurdrv()");
+		}
 	}
 #else	/* !MSDOS */
 	if (*path == _SC_) strcpy(resolved, _SS_);
 # ifndef	_NODOSDRIVE
 	else if ((drv = _dospath(path))) {
-		char cwd[MAXPATHLEN];
-
 		path += 2;
-		resolved[0] = drv;
-		resolved[1] = ':';
-		resolved[2] = '\0';
-		if (*path == _SC_ || !Xgetwd(cwd)
-		|| Xchdir(resolved) < 0 || !Xgetwd(resolved)) {
+		if (*path == _SC_) cp = NULL;
+		else {
+			duplastdrive = lastdrive;
+			lastdrive = drv;
+			cp = dosgetcwd(resolved, MAXPATHLEN - 1);
+			lastdrive = duplastdrive;
+		}
+		if (!cp) {
+			resolved[0] = drv;
+			resolved[1] = ':';
 			resolved[2] = _SC_;
 			resolved[3] = '\0';
 		}
-		else if (Xchdir(cwd) < 0) error(cwd);
 	}
 # endif
 #endif	/* !MSDOS */
-	else if (!*fullpath) {
-		if (!Xgetwd(resolved)) strcpy(resolved, _SS_);
-	}
-	else if (resolved != fullpath) strcpy(resolved, fullpath);
-	return(_realpath2(path, resolved, rdlink));
+	else if (!rdlink && resolved != fullpath && *fullpath)
+		strcpy(resolved, fullpath);
+	else if (!Xgetwd(resolved)) strcpy(resolved, _SS_);
+	norealpath++;
+	_realpath2(path, resolved, rdlink);
+	norealpath--;
+	return(resolved);
 }
 
 int _chdir2(path)
@@ -231,40 +280,72 @@ char *path;
 
 	if (!Xgetwd(cwd)) strcpy(cwd, _SS_);
 	if (Xchdir(path) < 0) return(-1);
-#if	!MSDOS && !defined (_NODOSDRIVE)
-	if (!dospath2("")) {
-		int fd;
+#if	!MSDOS
+# ifndef	_NODOSDRIVE
+	if (dospath2("")) /*EMPTY*/;
+	else
+# endif
+	{
+		int fd, duperrno;
 
 		if ((fd = open(".", O_RDONLY, 0600)) < 0) {
+# ifdef	CYGWIN
+			char upath[MAXPATHLEN], spath[MAXPATHLEN];
+			char tmp[MAXPATHLEN];
+			int len;
+
+			if (Xgetwd(tmp)) {
+				len = strlen(tmp);
+				cygwin_internal(CW_GET_CYGDRIVE_PREFIXES,
+					upath, spath);
+				if ((*upath && !strnpathcmp(tmp, upath, len)
+				&& (!upath[len] || upath[len] == _SC_))
+				|| (*spath && !strnpathcmp(tmp, spath, len)
+				&& (!spath[len] || spath[len] == _SC_)))
+					return(0);
+			}
+# endif	/* CYGWIN */
+			duperrno = errno;
 			if (Xchdir(cwd) < 0) lostcwd(cwd);
+			errno = duperrno;
 			return(-1);
 		}
 		close(fd);
 	}
-#endif
+#endif	/* !MSDOS */
 	return(0);
 }
 
 int chdir2(path)
 char *path;
 {
-	char *pwd, cwd[MAXPATHLEN], tmp[MAXPATHLEN];
+	char cwd[MAXPATHLEN], tmp[MAXPATHLEN];
+	int duperrno;
 
-	realpath2(path, tmp, 0);
-	if (_chdir2(path) < 0) return(-1);
+#ifdef	DEBUG
+	if (!path) {
+		if (lastpath) free(lastpath);
+		lastpath = NULL;
+# ifndef	_NODOSDRIVE
+		if (unixpath) free(unixpath);
+		unixpath = NULL;
+# endif
+		return(0);
+	}
+#endif
 
+	realpath2(path, tmp, physical_path);
 	strcpy(cwd, fullpath);
 	strcpy(fullpath, tmp);
 
 	if (_chdir2(fullpath) < 0) {
+		duperrno = errno;
 		if (_chdir2(cwd) < 0) lostcwd(fullpath);
 		else strcpy(fullpath, cwd);
+		errno = duperrno;
 		return(-1);
 	}
 	if (lastpath) free(lastpath);
-#ifdef	DEBUG
-	_mtrace_file = "chdir2";
-#endif
 	lastpath = strdup2(cwd);
 #ifndef	_NODOSDRIVE
 	if (dospath2(fullpath)) {
@@ -280,17 +361,9 @@ char *path;
 		unixpath = NULL;
 	}
 #endif
-	if (getenv("PWD")) {
-#ifdef	DEBUG
-		_mtrace_file = "chdir2(PWD)";
-#endif
-		pwd = malloc2(strlen(fullpath) + 4 + 1);
-		strcpy(strcpy2(pwd, "PWD="), fullpath);
-		if (putenv2(pwd) < 0) error("PWD");
-	}
+	if (getconstvar("PWD")) setenv2("PWD", fullpath, 1);
 #if	MSDOS
-	unixrealpath(fullpath, tmp);
-	strcpy(fullpath, tmp);
+	if (unixrealpath(fullpath, tmp)) strcpy(fullpath, tmp);
 #endif
 	entryhist(1, fullpath, 1);
 	return(0);
@@ -299,10 +372,10 @@ char *path;
 int chdir3(path)
 char *path;
 {
-	char *cwd;
 #ifndef	_NODOSDRIVE
 	int drive;
 #endif
+	char *cwd;
 
 	cwd = path;
 	if (!strcmp(path, ".")) cwd = NULL;
@@ -343,7 +416,7 @@ int mode;
 
 	eol = path + (int)strlen(path) - 1;
 	while (eol > path && *eol == _SC_) eol--;
-#if	MSDOS
+#ifdef	BSPATHDELIM
 	if (onkanji1(path, eol - path)) eol++;
 #endif
 	*(++eol) = '\0';
@@ -374,7 +447,12 @@ ALLOC_T size;
 {
 	char *tmp;
 
-	if (!size || !(tmp = (char *)malloc(size))) error(NULL);
+	if (!size || !(tmp = (char *)malloc(size))) {
+		error("malloc()");
+#ifdef	FAKEUNINIT
+		tmp = NULL;	/* fake for -Wuninitialized */
+#endif
+	}
 	return(tmp);
 }
 
@@ -386,18 +464,52 @@ ALLOC_T size;
 
 	if (!size
 	|| !(tmp = (ptr) ? (char *)realloc(ptr, size) : (char *)malloc(size)))
-		error(NULL);
+	{
+		error("realloc()");
+#ifdef	FAKEUNINIT
+		tmp = NULL;	/* fake for -Wuninitialized */
+#endif
+	}
 	return(tmp);
 }
 
+char *c_realloc(ptr, n, sizep)
+char *ptr;
+ALLOC_T n, *sizep;
+{
+	if (!ptr) {
+		*sizep = BUFUNIT;
+		return(malloc2(*sizep));
+	}
+	while (n + 1 >= *sizep) *sizep *= 2;
+	return(realloc2(ptr, *sizep));
+}
+
 char *strdup2(s)
 char *s;
 {
 	char *tmp;
+	int n;
 
 	if (!s) return(NULL);
-	if (!(tmp = (char *)malloc((ALLOC_T)strlen(s) + 1))) error(NULL);
-	strcpy(tmp, s);
+	n = strlen(s);
+	if (!(tmp = (char *)malloc((ALLOC_T)n + 1))) error("malloc()");
+	memcpy(tmp, s, n + 1);
+	return(tmp);
+}
+
+char *strndup2(s, n)
+char *s;
+int n;
+{
+	char *tmp;
+	int i;
+
+	if (!s) return(NULL);
+	for (i = 0; i < n; i++) if (!s[i]) break;
+	if (!(tmp = (char *)malloc((ALLOC_T)i + 1))) error("malloc()");
+	memcpy(tmp, s, i);
+	tmp[i] = '\0';
 	return(tmp);
 }
 
@@ -447,9 +559,13 @@ int n;
 
 	for (i = 0; i < n && s2[i]; i++) s1[i] = s2[i];
 	s1[i] = '\0';
-	return(s1);
+	return(&(s1[i]));
 }
 
+/*
+ *	strncpy3(buf, s, &(x), 0): same as sprintf(buf, "%-*.*s", x, x, s);
+ *	strncpy3(buf, s, &(-x), 0): same as sprintf(buf, "%s", s);
+ */
 int strncpy3(s1, s2, lenp, ptr)
 char *s1, *s2;
 int *lenp, ptr;
@@ -479,11 +595,11 @@ int *lenp, ptr;
 			s1[i++] = s2[j++];
 		}
 #else
-		if (iskna(s2[j]));
+		if (iskna(s2[j])) /*EMPTY*/;
 #endif
 		else if (iskanji1(s2, j)) {
 			if (*lenp >= 0 && i >= *lenp - 1) {
-				s1[i++] = ' ';
+				if (ptr >= 0) s1[i++] = ' ';
 				break;
 			}
 			s1[i++] = s2[j++];
@@ -516,18 +632,6 @@ int *lenp, ptr;
 	return(l);
 }
 
-char *strstr2(s1, s2)
-char *s1, *s2;
-{
-	char *cp;
-	int len;
-
-	len = strlen(s2);
-	for (cp = s1; (cp = strchr(cp, *s2)); cp++)
-		if (!strncmp(cp, s2, len)) return(cp);
-	return(NULL);
-}
-
 int strlen2(s)
 char *s;
 {
@@ -564,6 +668,81 @@ char *s;
 	return((int)n);
 }
 
+#ifdef	USESTDARGH
+/*VARARGS2*/
+int fprintf2(FILE *fp, CONST char *fmt, ...)
+#else
+/*VARARGS2*/
+int fprintf2(fp, fmt, va_alist)
+FILE *fp;
+CONST char *fmt;
+va_dcl
+#endif
+{
+	va_list args;
+	char *buf;
+	int n;
+
+#ifdef	USESTDARGH
+	va_start(args, fmt);
+#else
+	va_start(args);
+#endif
+
+	n = vasprintf2(&buf, fmt, args);
+	va_end(args);
+	if (n < 0) error("malloc()");
+	fputs(buf, fp);
+	free(buf);
+	return(n);
+}
+
+#ifdef	USESTDARGH
+/*VARARGS3*/
+int snprintf2(char *s, int size, CONST char *fmt, ...)
+#else
+/*VARARGS3*/
+int snprintf2(s, size, fmt, va_alist)
+char *s;
+int size;
+CONST char *fmt;
+va_dcl
+#endif
+{
+	va_list args;
+	char *buf;
+	int n;
+
+#ifdef	USESTDARGH
+	va_start(args, fmt);
+#else
+	va_start(args);
+#endif
+
+	n = vasprintf2(&buf, fmt, args);
+	va_end(args);
+	if (n < 0) error("malloc()");
+	strncpy2(s, buf, size - 1);
+	free(buf);
+	return(n);
+}
+
+VOID perror2(s)
+char *s;
+{
+	int duperrno;
+
+	duperrno = errno;
+	if (s) {
+		kanjifputs(s, stderr);
+		fputs(": ", stderr);
+	}
+	fputs(strerror2(duperrno), stderr);
+	if (isttyiomode) fputc('\r', stderr);
+	fputc('\n', stderr);
+	fflush(stderr);
+}
+
 #ifdef	_NOORIGSHELL
 static int NEAR _getenv2(name, len, envp)
 char *name;
@@ -575,7 +754,7 @@ char **envp;
 	if (!envp) return(-1);
 
 	for (i = 0; envp[i]; i++)
-		if (!strnpathcmp(envp[i], name, len) && envp[i][len] == '=')
+		if (!strnenvcmp(envp[i], name, len) && envp[i][len] == '=')
 			break;
 	return(i);
 }
@@ -583,7 +762,7 @@ char **envp;
 static char **NEAR _putenv2(s, envp)
 char *s, **envp;
 {
-	char *cp, *tmp, **new;
+	char *cp, **new;
 	int i, n, len;
 
 	if ((cp = strchr(s, '='))) len = (int)(cp - s);
@@ -591,104 +770,82 @@ char *s, **envp;
 
 	if ((n = _getenv2(s, len, envp)) < 0) n = 0;
 	else if (envp[n]) {
-		tmp = envp[n];
+		free(envp[n]);
 		if (cp) envp[n] = s;
 		else for (i = n; envp[i]; i++) envp[i] = envp[i + 1];
-		free(tmp);
 		return(envp);
 	}
 	if (!cp) return(envp);
 
-# ifdef	DEBUG
-	_mtrace_file = "_putenv2";
-# endif
-	if (!envp) new = (char **)malloc((n + 2) * sizeof(char *));
-	else new = (char **)realloc(envp, (n + 2) * sizeof(char *));
-	if (!new) {
-		free(s);
-		return(NULL);
-	}
+	new = (char **)realloc2(envp, (n + 2) * sizeof(char *));
 	new[n] = s;
 	new[n + 1] = (char *)NULL;
 	return(new);
 }
 #endif	/* _NOORIGSHELL */
 
-int putenv2(s)
-char *s;
-{
-#ifdef	_NOORIGSHELL
-	char **new;
-# if	MSDOS
-	char *cp;
-
-	for (cp = s; *cp && *cp != '='; cp++) *cp = toupper2(*cp);
-# endif
-	if (!(new = _putenv2(s, environ))) return(-1);
-	environ = new;
-	return(0);
-#else	/* !_NOORIGSHELL */
-	return(putexportvar(s, -1));
-#endif	/* !_NOORIGSHELL */
-}
-
 char *getenv2(name)
 char *name;
 {
 #ifdef	_NOORIGSHELL
 	char **envpp[2];
-	int i, c, n, len;
+	int i, n, len;
 
 	len = strlen(name);
-	c = strnpathcmp(name, "FD_", 3);
 	envpp[0] = environ2;
 	envpp[1] = environ;
 
 	for (i = 0; i < 2; i++) {
 		n = _getenv2(name, len, envpp[i]);
 		if (n >= 0 && envpp[i][n]) return(&(envpp[i][n][len + 1]));
-		if (c) continue;
-		n = _getenv2(name + 3, len - 3, envpp[i]);
-		if (n >= 0 && envpp[i][n]) return(&(envpp[i][n][len - 3 + 1]));
+		if (strnenvcmp(name, FDENV, FDESIZ)) continue;
+		n = _getenv2(&(name[FDESIZ]), len - FDESIZ, envpp[i]);
+		if (n >= 0 && envpp[i][n])
+			return(&(envpp[i][n][len - FDESIZ + 1]));
 	}
 #else	/* !_NOORIGSHELL */
 	char *cp;
 
 	if ((cp = getshellvar(name, -1))) return(cp);
-	if (!strnpathcmp(name, "FD_", 3) && (cp = getshellvar(name + 3, -1)))
+	if (!strnenvcmp(name, FDENV, FDESIZ)
+	&& (cp = getshellvar(&(name[FDESIZ]), -1)))
 		return(cp);
 #endif	/* !_NOORIGSHELL */
 	return(NULL);
 }
 
-int setenv2(name, value)
+int setenv2(name, value, export)
 char *name, *value;
+int export;
 {
 	char *cp;
 	int len;
-#ifdef	_NOORIGSHELL
-	char **new;
-# if	MSDOS
+#if	defined (ENVNOCASE) && defined (_NOORIGSHELL)
 	int i;
-# endif
 #endif
 
 	len = strlen(name);
-	if (!value) cp = name;
+	if (!value) {
+#ifdef	_NOORIGSHELL
+		cp = name;
+#else
+		return(unset(name, len));
+#endif
+	}
 	else {
 		cp = malloc2(len + strlen(value) + 2);
 		memcpy(cp, name, len);
-#if	MSDOS && defined (_NOORIGSHELL)
+#if	defined (ENVNOCASE) && defined (_NOORIGSHELL)
 		for (i = 0; i < len ; i++) cp[i] = toupper2(cp[i]);
 #endif
 		cp[len] = '=';
 		strcpy(&(cp[len + 1]), value);
 	}
 #ifdef	_NOORIGSHELL
-	if (!(new = _putenv2(cp, environ2))) return(-1);
-	environ2 = new;
+	if (export) environ = _putenv2(cp, environ);
+	else environ2 = _putenv2(cp, environ2);
 #else	/* !_NOORIGSHELL */
-	if (putshellvar(cp, len) < 0) {
+	if (((export) ? putexportvar(cp, len) : putshellvar(cp, len)) < 0) {
 		free(cp);
 		return(-1);
 	}
@@ -696,56 +853,132 @@ char *name, *value;
 	return(0);
 }
 
-#if	defined (DEBUG) && defined (_NOORIGSHELL)
-VOID freeenv(VOID_A)
+#ifdef	USESIGACTION
+sigcst_t signal2(sig, func)
+int sig;
+sigcst_t func;
 {
-	int i;
+	struct sigaction act, oact;
 
-	if (environ) {
-		for (i = 0; environ[i]; i++) free(environ[i]);
-		free(environ);
-	}
-	if (environ2) {
-		for (i = 0; environ2[i]; i++) free(environ2[i]);
-		free(environ2);
-	}
+	act.sa_handler = func;
+	act.sa_flags = 0;
+	sigemptyset(&(act.sa_mask));
+	sigemptyset(&(oact.sa_mask));
+	if (sigaction(sig, &act, &oact) < 0) return(SIG_ERR);
+	return(oact.sa_handler);
 }
-#endif	/* DEBUG && _NOORIGSHELL */
+#endif	/* USESIGACTION */
 
 int system2(command, noconf)
 char *command;
 int noconf;
 {
-	int n, ret;
+	int n, wastty, mode, nl, ret;
 
 	if (!command || !*command) return(0);
+#ifdef	FAKEUNINIT
+	mode = nl = 0;		/* fake for -Wuninitialized */
+#endif
 	n = sigvecset(0);
-	if (noconf >= 0) {
-		locate(0, n_line - 1);
-		putterm(l_clear);
-		if (n && noconf) putterms(t_end);
-	}
-	stdiomode();
-	ret = dosystem(command);
-	if (noconf > 0) {
-		if (ret >= 127) {
-			fputs("\007\n", stderr);
-			kanjifputs(HITKY_K, stderr);
-			fflush(stderr);
-			ttyiomode();
-			keyflush();
-			getkey2(0);
-			stdiomode();
-			fputc('\n', stderr);
+	if ((wastty = isttyiomode)) {
+		if (noconf >= 0) {
+			locate(0, n_line - 1);
+			putterm(l_clear);
 		}
-		if (n) putterms(t_init);
+		if (n && noconf > 0) mode = termmode(0);
+		nl = stdiomode();
 	}
-	ttyiomode();
+
+	ret = dosystem(command);
 	sigvecset(n);
-	if (!noconf || (noconf < 0 && ret >= 127)) {
-		hideclock = 1;
-		warning(0, HITKY_K);
+	if (ret >= 127 && noconf > 0) {
+		if (dumbterm <= 2) fputc('\007', stderr);
+		fputc('\n', stderr);
+		kanjifputs(HITKY_K, stderr);
+		fflush(stderr);
+		ttyiomode(1);
+		keyflush();
+		getkey2(0);
+		stdiomode();
+		fputc('\n', stderr);
 	}
+
+	if (wastty) {
+		ttyiomode(nl);
+		if (n && noconf > 0) termmode(mode);
+		if (!noconf || (noconf < 0 && ret >= 127)) {
+			hideclock = 1;
+			warning(0, HITKY_K);
+		}
+	}
+	return(ret);
+}
+
+/*ARGSUSED*/
+FILE *popen2(command, type)
+char *command, *type;
+{
+	FILE *fp;
+	int n;
+
+	if (!command || !*command) return(NULL);
+	n = sigvecset(0);
+	wasttyflags = 0;
+	if (isttyiomode) {
+		wasttyflags |= TF_TTYIOMODE;
+		if (stdiomode()) wasttyflags |= TF_TTYNL;
+	}
+
+#ifdef	_NOORIGSHELL
+# ifdef	DEBUG
+	_mtrace_file = "popen(start)";
+	fp = Xpopen(command, type);
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "popen(end)";
+		malloc(0);	/* dummy malloc */
+	}
+# else
+	fp = Xpopen(command, type);
+# endif
+#else	/* !_NOORIGSHELL */
+	fp = dopopen(command);
+#endif	/* !_NOORIGSHELL */
+	sigvecset(n);
+	if (fp) {
+		if (wasttyflags & TF_TTYIOMODE) {
+			putterms(t_keypad);
+			tflush();
+		}
+	}
+	else {
+		if (dumbterm <= 2) fputc('\007', stderr);
+		fputc('\n', stderr);
+		perror2(command);
+		fflush(stderr);
+		ttyiomode(1);
+		keyflush();
+		getkey2(0);
+		stdiomode();
+		fputc('\n', stderr);
+		if (wasttyflags & TF_TTYIOMODE)
+			ttyiomode((wasttyflags & TF_TTYNL) ? 1 : 0);
+	}
+	return(fp);
+}
+
+int pclose2(fp)
+FILE *fp;
+{
+	int ret;
+
+#ifdef	_NOORIGSHELL
+	ret = Xpclose(fp);
+#else
+	ret = dopclose(fp);
+#endif
+	if (wasttyflags & TF_TTYIOMODE)
+		ttyiomode((wasttyflags & TF_TTYNL) ? 1 : 0);
 	return(ret);
 }
 
@@ -753,7 +986,7 @@ char *getwd2(VOID_A)
 {
 	char cwd[MAXPATHLEN];
 
-	if (!Xgetwd(cwd)) error(NULL);
+	if (!Xgetwd(cwd)) lostcwd(cwd);
 	return(strdup2(cwd));
 }
 
@@ -781,19 +1014,19 @@ static long NEAR gettimezone(tm, t)
 struct tm *tm;
 time_t t;
 {
-#if	MSDOS
+# if	MSDOS
 	struct timeb buffer;
 
 	ftime(&buffer);
 	return((long)(buffer.timezone) * 60L);
-#else	/* !MSDOS */
-#ifdef	NOTMGMTOFF
+# else	/* !MSDOS */
+#  ifdef	NOTMGMTOFF
 	struct timeval t_val;
 	struct timezone t_zone;
-#endif
-#ifndef	NOTZFILEH
+#  endif
+#  ifndef	NOTZFILEH
 	struct tzhead head;
-#endif
+#  endif
 	struct tm tmbuf;
 	FILE *fp;
 	time_t tmp;
@@ -803,15 +1036,15 @@ time_t t;
 
 	memcpy((char *)&tmbuf, (char *)tm, sizeof(struct tm));
 
-#ifdef	NOTMGMTOFF
+#  ifdef	NOTMGMTOFF
 	gettimeofday2(&t_val, &t_zone);
 	tz = t_zone.tz_minuteswest * 60L;
-#else
+#  else
 	tz = -(localtime(&t) -> tm_gmtoff);
-#endif
+#  endif
 
-#ifndef	NOTZFILEH
-	cp = (char *)getenv("TZ");
+#  ifndef	NOTZFILEH
+	cp = getconstvar("TZ");
 	if (!cp || !*cp) cp = TZDEFAULT;
 	if (cp[0] == _SC_) strcpy(buf, cp);
 	else strcatdelim2(buf, TZDIR, cp);
@@ -820,11 +1053,11 @@ time_t t;
 		fclose(fp);
 		return(tz);
 	}
-#ifdef	USELEAPCNT
+#   ifdef	USELEAPCNT
 	nleap = char2long(head.tzh_leapcnt);
-#else
+#   else
 	nleap = char2long(head.tzh_timecnt - 4);
-#endif
+#   endif
 	ntime = char2long(head.tzh_timecnt);
 	ntype = char2long(head.tzh_typecnt);
 	nchar = char2long(head.tzh_charcnt);
@@ -882,23 +1115,39 @@ time_t t;
 
 	tz += leap;
 	fclose(fp);
-#endif	/* !NOTZFILEH */
+#  endif	/* !NOTZFILEH */
 
 	return(tz);
-#endif	/* !MSDOS */
+# endif	/* !MSDOS */
 }
 #endif	/* !USEMKTIME && !USETIMELOCAL */
 
+time_t time2(VOID_A)
+{
+#if	MSDOS
+	struct timeb buffer;
+
+	ftime(&buffer);
+	return((time_t)(buffer.time));
+#else
+	struct timeval t_val;
+	struct timezone tz;
+
+	gettimeofday2(&t_val, &tz);
+	return((time_t)(t_val.tv_sec));
+#endif
+}
+
 time_t timelocal2(tm)
 struct tm *tm;
 {
 #ifdef	USEMKTIME
 	tm -> tm_isdst = -1;
 	return(mktime(tm));
-#else
+#else	/* !USEMKTIME */
 # ifdef	USETIMELOCAL
 	return(timelocal(tm));
-# else
+# else	/* !USETIMELOCAL */
 	time_t d, t;
 	int i, y;
 
@@ -933,8 +1182,8 @@ struct tm *tm;
 	t += gettimezone(tm, t);
 
 	return(t);
-# endif
-#endif
+# endif	/* !USETIMELOCAL */
+#endif	/* !USEMKTIME */
 }
 
 char *fgets2(fp, nulcnv)
@@ -942,10 +1191,10 @@ FILE *fp;
 int nulcnv;
 {
 	char *cp;
-	long i, size;
+	ALLOC_T i, size;
 	int c;
 
-	cp = c_malloc(size);
+	cp = c_realloc(NULL, 0, &size);
 	for (i = 0; (c = Xfgetc(fp)) != '\n'; i++) {
 		if (c == EOF) {
 			if (!i || ferror(fp)) {
@@ -954,11 +1203,11 @@ int nulcnv;
 			}
 			break;
 		}
-		cp = c_realloc(cp, i, size);
+		cp = c_realloc(cp, i, &size);
 		if (nulcnv && !c) c = '\n';
 		cp[i] = c;
 	}
-#if	MSDOS
+#ifdef	USECRNL
 	if (i > 0 && cp[i - 1] == '\r') i--;
 #endif
 	cp[i++] = '\0';
