@@ -5,6 +5,13 @@
  */
 
 #include "fd.h"
+#if	!MSDOS
+#include <sys/param.h>
+#endif
+
+#ifdef	USESELECTH
+#include <sys/select.h>
+#endif
 
 #ifdef	NOVOID
 #define	VOID
@@ -12,6 +19,24 @@
 #define	VOID	void
 #endif
 
+#if	defined (USESYSCONF) && defined (_SC_OPEN_MAX)
+#define	MAXOPENFILE	sysconf(_SC_OPEN_MAX)
+#else
+# ifdef	NOFILE
+# define	MAXOPENFILE	NOFILE
+# else
+#  ifdef	OPEN_MAX
+#  define	MAXOPENFILE	OPEN_MAX
+#  else
+#   if	MSDOS
+#   define	MAXOPENFILE	20
+#   else
+#   define	MAXOPENFILE	64
+#   endif
+#  endif
+# endif
+#endif
+
 
 /*ARGSUSED*/
 int main(argc, argv)
@@ -22,22 +47,33 @@ char *argv[];
 	char *cp;
 #endif
 
-	printf("s:__VERSION__:%d:g\n", FD);
+	printf("s:__VERSION__:%d:\n", FD);
+/*NOTDEFINED*/
 #if	FD >= 2
-	printf("s:__RCVERSION__:%d:g\n", FD);
+	printf("s:__RCVERSION__:%d:\n", FD);
 #else
-	printf("s:__RCVERSION__::g\n");
+	printf("s:__RCVERSION__::\n");
 #endif
+/*NOTDEFINED*/
+	printf("s:__PREFIX__:%s:\n", PREFIX);
 	printf("s:__EXE__::g\n");
 	printf("s:__OBJ__:.o:g\n");
 	printf("s:__OBJS__:dosemu.o:\n");
 	printf("s:__OBJLIST__:$(OBJ1) $(OBJ2) $(OBJ3):\n");
-	printf("s:__SOBJLIST__:$(SOBJ):\n");
 	printf("s:__DEFRC__:'\"'$(DEFRC)'\"':\n");
 
+#ifdef	USEDATADIR
+	printf("s:__UNITBLPATH__:-DDATADIR='\"'$(DATADIR)'\"':\n");
+	printf("s:__DATADIR__:$(DATADIR):g\n");
+#else
+	printf("s:__UNITBLPATH__::\n");
+	printf("s:__DATADIR__:$(BINDIR):g\n");
+#endif
+
 	printf("s:	__RENAME__:#	mv:\n");
 	printf("s:	__AOUT2EXE__:#	aout2exe:\n");
 	printf("s:	__REMOVE__:#	rm -f:\n");
+	printf("s:__ECHO__:echo:\n");
 	printf("s:__COPY__:cp:\n");
 	printf("s:__RM__:rm -f:\n");
 
@@ -65,8 +101,26 @@ char *argv[];
 
 	printf("s:__CC__:%s:\n", CCCOMMAND);
 	printf("s:__CCOPTIONS__:%s:\n", EXTENDCCOPT);
+#ifdef	HOSTCCCOMMAND
+	if (strcmp(CCCOMMAND, HOSTCCCOMMAND)) {
+		printf("s:__HOSTCC__:%s:\n", HOSTCCCOMMAND);
+		printf("s:__HOSTCCOPTIONS__:-O:\n");
+	}
+	else
+#endif
+	{
+		printf("s:__HOSTCC__:$(CC):\n");
+		printf("s:__HOSTCCOPTIONS__:%s:\n", EXTENDCCOPT);
+	}
+#if	defined (DEFFDSETSIZE) && defined (FD_SETSIZE)
+	if (FD_SETSIZE < MAXOPENFILE)
+		printf("s:__FDSETSIZE__:-DFD_SETSIZE=%d:\n", MAXOPENFILE);
+	else
+#endif
+	printf("s:__FDSETSIZE__::\n");
 	printf("s:__MEM__::\n");
 	printf("s:__SHMEM__::\n");
+	printf("s:__BSHMEM__::\n");
 #ifdef	CCOUTOPT
 	printf("s:__OUT__:%s:\n", CCOUTOPT);
 #else
