@@ -4,7 +4,7 @@
  *	Tree Mode Module
  */
 
-#include <signal.h>
+#include <ctype.h>
 #include "fd.h"
 #include "term.h"
 #include "func.h"
@@ -12,13 +12,8 @@
 
 #ifndef	_NOTREE
 
-#if	MSDOS
-#include <ctype.h>
-# ifndef	_NODOSDRIVE
+#if	MSDOS && !defined (_NODOSDRIVE)
 extern int preparedrv __P_((int));
-# endif
-#else
-#include <sys/param.h>
 #endif
 
 
@@ -101,7 +96,7 @@ int disp;
 	if (!(dirp = Xopendir(dir))) return(0);
 	while ((dp = Xreaddir(dirp))) {
 		if (isdotdir(dp -> d_name)) continue;
-		strcpy(path + len, dp -> d_name);
+		strcpy(&(path[len]), dp -> d_name);
 		if (limit-- <= 0
 		|| (stat2(path, &st) >= 0
 		&& (st.st_mode & S_IFMT) == S_IFDIR)) {
@@ -138,8 +133,8 @@ int level, *maxp;
 	DIR *dirp;
 	struct dirent *dp;
 	struct stat st;
-	char *cp, *dir, *subdir;
-#if	!MSDOS && !defined (_NODOSDRIVE)
+	char *cp, *dir, *subdir, cwd[MAXPATHLEN];
+#ifdef	_USEDOSEMU
 	char tmp[MAXPATHLEN];
 #endif
 	int i, len;
@@ -152,6 +147,11 @@ int level, *maxp;
 		*maxp = -1;
 		return(NULL);
 	}
+	if (!Xgetwd(cwd)) {
+		*maxp = 0;
+		return(NULL);
+	}
+
 #if	MSDOS
 	if (_dospath(path)) path += 2;
 #endif
@@ -160,18 +160,16 @@ int level, *maxp;
 		subdir = path + 1;
 		if (!*subdir) subdir = NULL;
 	}
-#if	!MSDOS && !defined (_NODOSDRIVE)
+#ifdef	_USEDOSEMU
 	else if (_dospath(path)) {
-		dir = malloc2(3 + 1);
-		strncpy2(dir, path, 3);
+		dir = strndup2(path, 3);
 		subdir = path + 3;
 		if (!*subdir) subdir = NULL;
 	}
 #endif
 	else {
 		len = (cp = strdelim(path, 0)) ? cp - path : strlen(path);
-		dir = malloc2(len + 1);
-		strncpy2(dir, path, len);
+		dir = strndup2(path, len);
 		subdir = (cp) ? cp + 1 : NULL;
 	}
 
@@ -193,7 +191,7 @@ int level, *maxp;
 			list[*maxp].name = strdup2(dp -> d_name);
 			list[*maxp].sub = NULL;
 			list[*maxp].max = 0;
-#if	!MSDOS
+#ifndef	NODIRLOOP
 			list[*maxp].dev = st.st_dev;
 			list[*maxp].ino = st.st_ino;
 			list[*maxp].parent = parent;
@@ -205,29 +203,41 @@ int level, *maxp;
 			list[0].name = strdup2(dp -> d_name);
 			list[0].sub = &(list[0]);
 			list[0].max = 0;
-#if	!MSDOS
+#ifndef	NODIRLOOP
 			list[0].dev = st.st_dev;
 			list[0].ino = st.st_ino;
 			list[0].parent = parent;
 #endif
 			if (++(*maxp) >= 2) break;
 		}
-		else if (!i) {
-			i++;
-			list[1].name = NULL;
-			list[1].sub = NULL;
-			list[1].max = -1;
-#if	!MSDOS
-			list[1].dev = 0;
-			list[1].ino = 0;
-			list[1].parent = NULL;
+		else {
+			if (!(*maxp)) {
+				list[0].name = NULL;
+				list[0].sub = NULL;
+				list[0].max = -1;
+#ifndef	NODIRLOOP
+				list[0].dev = 0;
+				list[0].ino = 0;
+				list[0].parent = NULL;
 #endif
-			if (++(*maxp) >= 2) break;
+			}
+			if (!i) {
+				i++;
+				list[1].name = NULL;
+				list[1].sub = NULL;
+				list[1].max = -1;
+#ifndef	NODIRLOOP
+				list[1].dev = 0;
+				list[1].ino = 0;
+				list[1].parent = NULL;
+#endif
+				if (++(*maxp) >= 2) break;
+			}
 		}
 	}
 	Xclosedir(dirp);
 	if (list) for (i = 0; i < *maxp; i++) {
-#if	!MSDOS
+#ifndef	NODIRLOOP
 		treelist *lp;
 
 		if (!(list[i].name)) lp = NULL;
@@ -244,7 +254,7 @@ int level, *maxp;
 			list[i].sub = maketree(subdir, NULL, &(list[i]),
 				level + 1, &(list[i].max));
 		else {
-			if (list[i].max >= 0
+			if (list[i].max >= 0 && list[i].name
 			&& evaldir(nodospath(tmp, list[i].name), 0))
 				list[i].max = -1;
 		}
@@ -253,17 +263,16 @@ int level, *maxp;
 
 #if	MSDOS
 	if (_dospath(path)) path += 2;
-	if (*path == _SC_) {
-#else	/* !MSDOS */
-	if (*path == _SC_
-#ifndef	_NODOSDRIVE
-	|| _dospath(path)
 #endif
-	) {
-#endif	/* !MSDOS */
+#if	MSDOS || defined (_NODOSDRIVE)
+	if (*path == _SC_)
+#else
+	if (*path == _SC_ || _dospath(path))
+#endif
+	{
 		if (_chdir2(fullpath) < 0) error(fullpath);
 	}
-	else if (strcmp(path, ".") && _chdir2("..") < 0) error("..");
+	else if (strcmp(path, ".") && _chdir2(cwd) < 0) error("..");
 	return(list);
 }
 
@@ -315,8 +324,10 @@ static VOID NEAR showtree(VOID_A)
 		locate(0, i);
 		putterm(l_clear);
 	}
-	for (i = LFILETOP + 1; i < LFILEBOTTOM; i++)
-		sprintf(bufptr(i), "%-*.*s", TREEFIELD, TREEFIELD, " ");
+	for (i = LFILETOP + 1; i < LFILEBOTTOM; i++) {
+		memset(bufptr(i), ' ', TREEFIELD);
+		bufptr(i)[TREEFIELD] = '\0';
+	}
 	_showtree(tr_root -> sub, 1, 0, tr_top);
 	for (i = LFILETOP + 1; i < LFILEBOTTOM; i++) {
 		locate(1, i);
@@ -333,13 +344,13 @@ static VOID NEAR treebar(VOID_A)
 	locate(1, LFILETOP);
 	cputs2("Tree=");
 	kanjiputs2(treepath, n_column - 6, 0);
-	locate(0, LMESLINE);
+	locate(0, L_MESLINE);
 	putterm(l_clear);
 }
 
 VOID rewritetree(VOID_A)
 {
-	tr_scr = realloc2(tr_scr, (FILEPERLOW - 1) * (TREEFIELD + 1));
+	tr_scr = realloc2(tr_scr, (FILEPERROW - 1) * (TREEFIELD + 1));
 	searchtree();
 	showtree();
 	treebar();
@@ -375,18 +386,19 @@ int max, nest;
 				for (j = strlen(treepath); j >= 0; j--)
 					treepath[j + len] = treepath[j];
 				memcpy(treepath, list[i].name, len - 1);
-#else
+#else	/* !MSDOS */
+# ifdef	_NODOSDRIVE
+				if (*list[i].name != _SC_) len++;
+# else
 				if (*list[i].name != _SC_
-#ifndef	_NODOSDRIVE
-				&& !_dospath(list[i].name)
-#endif
-				) len++;
+				&& !_dospath(list[i].name)) len++;
+# endif
 				for (j = strlen(treepath); j >= 0; j--)
 					treepath[j + len] = treepath[j];
 				if (*list[i].name != _SC_)
 					memcpy(treepath, list[i].name,
 						len - 1);
-#endif
+#endif	/* !MSDOS */
 				treepath[len - 1] = _SC_;
 			}
 		}
@@ -413,7 +425,10 @@ treelist *list;
 		memcpy((char *)lp, (char *)&(list -> sub[0]),
 			sizeof(treelist));
 		for (i = 1; i < list -> max; i++)
-			if ((list -> sub[i]).name) free((list -> sub[i]).name);
+			if ((list -> sub[i]).name) {
+				free((list -> sub[i]).name);
+				(list -> sub[i]).name = NULL;
+			}
 	}
 
 	if (_chdir2(treepath) < 0) {
@@ -435,8 +450,10 @@ treelist *list;
 		return(i);
 	}
 	if (list -> sub) {
-		for (i = 0; i < list -> max; i++)
-			if (!strpathcmp(lp -> name, lptmp[i].name)) break;
+		if (!(lp -> name)) i = list -> max;
+		else for (i = 0; i < list -> max; i++)
+			if (lptmp[i].name
+			&& !strpathcmp(lp -> name, lptmp[i].name)) break;
 		if (i < list -> max) {
 			free(lptmp[i].name);
 			for (; i > 0; i--) memcpy((char *)&(lptmp[i]),
@@ -552,13 +569,9 @@ static int NEAR _tree_input(VOID_A)
 	int ch, tmp, half;
 
 	keyflush();
-#ifdef	_NOEDITMODE
-	ch = Xgetkey(SIGALRM, 0);
-#else
 	Xgetkey(-1, 0);
-	ch = Xgetkey(SIGALRM, 0);
+	ch = Xgetkey(1, 0);
 	Xgetkey(-1, 0);
-#endif
 
 	old = tr_cur;
 	switch (ch) {
@@ -569,7 +582,7 @@ static int NEAR _tree_input(VOID_A)
 			treedown();
 			break;
 		case K_PPAGE:
-			half = (FILEPERLOW - 1) / 2;
+			half = (FILEPERROW - 1) / 2;
 			tmp = LFILETOP + half + 1 - tr_line;
 			if (tr_top + half > LFILETOP + 1)
 				half = LFILETOP - tr_top + 1;
@@ -579,7 +592,7 @@ static int NEAR _tree_input(VOID_A)
 			searchtree();
 			break;
 		case K_NPAGE:
-			half = (FILEPERLOW - 1) / 2;
+			half = (FILEPERROW - 1) / 2;
 			tmp = half + (LFILETOP + half + 1) - tr_line;
 			while (tmp-- > 0) if (treedown() < 0) break;
 			tmp = tr_line - (LFILETOP + half + 1);
@@ -663,7 +676,7 @@ static int NEAR _tree_input(VOID_A)
 			ch = '\0';
 			free(cwd);
 			break;
-		case CTRL('L'):
+		case K_CTRL('L'):
 			rewritefile(1);
 			break;
 		case K_ESC:
@@ -686,7 +699,7 @@ static int NEAR _tree_input(VOID_A)
 
 static char *NEAR _tree(VOID_A)
 {
-#if	!MSDOS
+#ifndef	NODIRLOOP
 	struct stat st;
 # ifndef	_NODOSDRIVE
 	char tmp[MAXPATHLEN];
@@ -701,7 +714,7 @@ static char *NEAR _tree(VOID_A)
 	tr_root -> name = NULL;
 	tr_root -> max = 1;
 	tr_root -> sub = tr_cur = (treelist *)malloc2(sizeof(treelist));
-#if	!MSDOS
+#ifndef	NODIRLOOP
 	tr_root -> ino = 0;
 	tr_root -> parent = NULL;
 #endif
@@ -715,16 +728,15 @@ static char *NEAR _tree(VOID_A)
 	strcpy(path, fullpath);
 	tr_cur[0].name = strdup2(_SS_);
 # else
-	if (dospath("", path)) {
-		tr_cur[0].name = malloc2(3 + 1);
-		strncpy2(tr_cur[0].name, path, 3);
-	}
+	if (dospath("", path)) tr_cur[0].name = strndup2(path, 3);
 	else {
 		strcpy(path, fullpath);
 		tr_cur[0].name = strdup2(_SS_);
 	}
 # endif
+#endif	/* !MSDOS */
 
+#ifndef	NODIRLOOP
 	if (Xstat(nodospath(tmp, tr_cur[0].name), &st) < 0)
 		tr_cur[0].dev = tr_cur[0].ino = 0;
 	else {
@@ -732,7 +744,7 @@ static char *NEAR _tree(VOID_A)
 		tr_cur[0].ino = st.st_ino;
 	}
 	tr_cur[0].parent = NULL;
-#endif	/* !MSDOS */
+#endif
 	tr_cur[0].sub = maketree(path, NULL, &(tr_cur[0]),
 		0, &(tr_cur[0].max));
 
@@ -741,11 +753,11 @@ static char *NEAR _tree(VOID_A)
 	cwd = path + 2;
 #else	/* !MSDOS */
 	cwd = path;
-#ifndef	_NODOSDRIVE
+# ifndef	_NODOSDRIVE
 	if (_dospath(cwd)) cwd += 2;
-#endif
+# endif
 #endif	/* !MSDOS */
-	if (!strcmp(cwd, _SS_));
+	if (!strcmp(cwd, _SS_)) /*EMPTY*/;
 	else for (cp = cwd; (cp = strdelim(cp, 0)); cp++, tr_line++)
 		if ((tr_line + 1) * DIRFIELD + 2 > TREEFIELD
 		|| !(tr_cur = &(tr_cur -> sub[0]))) break;
@@ -755,7 +767,7 @@ static char *NEAR _tree(VOID_A)
 		tr_line = LFILEBOTTOM - 2;
 	}
 
-	tr_scr = malloc2((FILEPERLOW - 1) * (TREEFIELD + 1));
+	tr_scr = malloc2((FILEPERROW - 1) * (TREEFIELD + 1));
 	searchtree();
 	showtree();
 	win_x = 0;
@@ -800,6 +812,10 @@ int cleanup, *ddp;
 {
 	char *path, *dupfullpath;
 
+	if (FILEPERROW < WFILEMINTREE) {
+		warning(0, NOROW_K);
+		return(NULL);
+	}
 	dupfullpath = strdup2(fullpath);
 	do {
 		path = _tree();
